@@ -6,7 +6,8 @@
  */
 
 #include "icon.h"
-#include "scenegraph/managedtexturenode.h"
+#include "scenegraph/shadernode.h"
+#include "scenegraph/softwarerectanglenode.h"
 
 #include "platform/platformtheme.h"
 #include "platform/units.h"
@@ -20,12 +21,10 @@
 #include <QPropertyAnimation>
 #include <QQuickImageProvider>
 #include <QQuickWindow>
-#include <QSGSimpleTextureNode>
 #include <QSGTexture>
 #include <QScreen>
-#include <cstdlib>
 
-Q_GLOBAL_STATIC(ImageTexturesCache, s_iconImageCache)
+using namespace Qt::StringLiterals;
 
 Icon::Icon(QQuickItem *parent)
     : QQuickItem(parent)
@@ -39,7 +38,6 @@ Icon::Icon(QQuickItem *parent)
 
     connect(this, &QQuickItem::smoothChanged, this, &QQuickItem::polish);
     connect(this, &QQuickItem::enabledChanged, this, [this]() {
-        m_allowNextAnimation = true;
         polish();
     });
 }
@@ -109,9 +107,6 @@ void Icon::setActive(const bool active)
         return;
     }
     m_active = active;
-    if (isComponentComplete()) {
-        m_allowNextAnimation = true;
-    }
     polish();
     Q_EMIT activeChanged();
 }
@@ -179,105 +174,69 @@ QColor Icon::color() const
     return m_color;
 }
 
-QSGNode *Icon::createSubtree(qreal initialOpacity)
-{
-    auto opacityNode = new QSGOpacityNode{};
-    opacityNode->setFlag(QSGNode::OwnedByParent, true);
-    opacityNode->setOpacity(initialOpacity);
-
-    auto *mNode = new ManagedTextureNode;
-
-    mNode->setTexture(s_iconImageCache->loadTexture(window(), m_icon, QQuickWindow::TextureCanUseAtlas));
-
-    opacityNode->appendChildNode(mNode);
-
-    return opacityNode;
-}
-
-void Icon::updateSubtree(QSGNode *node, qreal opacity)
-{
-    auto opacityNode = static_cast<QSGOpacityNode *>(node);
-    opacityNode->setOpacity(opacity);
-
-    auto textureNode = static_cast<ManagedTextureNode *>(opacityNode->firstChild());
-    textureNode->setFiltering(smooth() ? QSGTexture::Linear : QSGTexture::Nearest);
-}
-
 QSGNode *Icon::updatePaintNode(QSGNode *node, QQuickItem::UpdatePaintNodeData * /*data*/)
 {
     if (m_source.isNull() || qFuzzyIsNull(width()) || qFuzzyIsNull(height())) {
         delete node;
-        return Q_NULLPTR;
+        return nullptr;
     }
 
-    if (!node) {
-        node = new QSGNode{};
-    }
+    m_sizeChanged = false;
 
-    if (m_animation && m_animation->state() == QAbstractAnimation::Running) {
-        if (node->childCount() < 2) {
-            node->appendChildNode(createSubtree(0.0));
-            m_textureChanged = true;
+    if (isSoftwareRendering()) {
+        auto rectangleNode = static_cast<SoftwareRectangleNode *>(node);
+        if (!rectangleNode) {
+            rectangleNode = new SoftwareRectangleNode{};
         }
 
-        // Rather than doing a perfect crossfade, first fade in the new texture
-        // then fade out the old texture. This is done to avoid the underlying
-        // color bleeding through when both textures are at ~0.5 opacity, which
-        // causes flickering if the two textures are very similar.
-        updateSubtree(node->firstChild(), 2.0 - m_animValue * 2.0);
-        updateSubtree(node->lastChild(), m_animValue * 2.0);
+        rectangleNode->setWindow(window());
+        rectangleNode->setRect(calculateNodeRect());
+        rectangleNode->setImage(m_icon);
+        return rectangleNode;
+    }
+
+    auto shaderNode = static_cast<ShaderNode *>(node);
+    if (!shaderNode) {
+        shaderNode = new ShaderNode{};
+    }
+
+    QString shaderName = u"icon_"_s;
+    if (isMask()) {
+        shaderName += u"mask_"_s;
+    }
+    if (m_animated) {
+        shaderName += u"mix"_s;
     } else {
-        if (node->childCount() == 0) {
-            node->appendChildNode(createSubtree(1.0));
-            m_textureChanged = true;
-        }
+        shaderName += u"default"_s;
+    }
+    shaderNode->setShader(shaderName);
+    shaderNode->setUniformBufferSize(sizeof(float) * 24);
 
-        if (node->childCount() > 1) {
-            auto toRemove = node->firstChild();
-            node->removeChildNode(toRemove);
-            delete toRemove;
-        }
-
-        updateSubtree(node->firstChild(), 1.0);
+    if (m_animated) {
+        shaderNode->setTextureChannels(2);
+    } else {
+        shaderNode->setTextureChannels(1);
     }
 
-    if (m_textureChanged) {
-        auto mNode = static_cast<ManagedTextureNode *>(node->lastChild()->firstChild());
-        mNode->setTexture(s_iconImageCache->loadTexture(window(), m_icon, QQuickWindow::TextureCanUseAtlas));
-        m_textureChanged = false;
-        m_sizeChanged = true;
+    auto maskColor = !m_color.isValid() || m_color == Qt::transparent ? (m_selected ? m_theme->highlightedTextColor() : m_theme->textColor()) : m_color;
+
+    UniformDataStream stream(shaderNode->uniformData());
+    stream.skipMatrixOpacity();
+    stream << float(m_animValue) // mix_amount
+           << (m_active ? 0.7f : 0.0f) // highlight_amount
+           << (isEnabled() ? 0.0f : 1.0f) // desaturate_amount
+           << ShaderNode::toPremultiplied(maskColor); // mask_color
+
+    shaderNode->setTexture(0, m_icon, window(), QQuickWindow::TextureCanUseAtlas);
+
+    if (!m_oldIcon.isNull() && m_animated) {
+        shaderNode->setTexture(1, m_oldIcon, window(), QQuickWindow::TextureCanUseAtlas);
     }
 
-    if (m_sizeChanged) {
-        const QSizeF iconPixSize(m_icon.width() / m_devicePixelRatio, m_icon.height() / m_devicePixelRatio);
-        const QSizeF itemPixSize = QSizeF((size() * m_devicePixelRatio).toSize()) / m_devicePixelRatio;
-        QRectF nodeRect(QPoint(0, 0), itemPixSize);
+    shaderNode->setRect(calculateNodeRect());
+    shaderNode->update();
 
-        if (itemPixSize.width() != 0 && itemPixSize.height() != 0) {
-            if (iconPixSize != itemPixSize) {
-                // At this point, the image will already be scaled, but we need to output it in
-                // the correct aspect ratio, painted centered in the viewport. So:
-                QRectF destination(QPointF(0, 0), QSizeF(m_icon.size()).scaled(m_paintedSize, Qt::KeepAspectRatio));
-                destination.moveCenter(nodeRect.center());
-                destination.moveTopLeft(QPointF(destination.topLeft().toPoint() * m_devicePixelRatio) / m_devicePixelRatio);
-                nodeRect = destination;
-            }
-        }
-
-        // Adjust the final node on the pixel grid
-        QPointF globalPixelPos = mapToScene(nodeRect.topLeft()) * m_devicePixelRatio;
-        QPointF posAdjust = QPointF(globalPixelPos.x() - std::round(globalPixelPos.x()), globalPixelPos.y() - std::round(globalPixelPos.y()));
-        nodeRect.moveTopLeft(nodeRect.topLeft() - posAdjust);
-
-        for (int i = 0; i < node->childCount(); ++i) {
-            auto mNode = static_cast<ManagedTextureNode *>(node->childAtIndex(i)->firstChild());
-            mNode->setRect(nodeRect);
-        }
-
-        m_sizeChanged = false;
-    }
-
-    return node;
+    return shaderNode;
 }
 
 void Icon::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
@@ -393,30 +352,16 @@ void Icon::updatePolish()
             m_icon = QImage(size, QImage::Format_Alpha8);
             m_icon.fill(Qt::transparent);
         }
-
-        const QColor tintColor = //
-            !m_color.isValid() || m_color == Qt::transparent //
-            ? (m_selected ? m_theme->highlightedTextColor() : m_theme->textColor())
-            : m_color;
-
-        // TODO: initialize m_isMask with icon.isMask()
-        if (tintColor.alpha() > 0 && isMask()) {
-            QPainter p(&m_icon);
-            p.setCompositionMode(QPainter::CompositionMode_SourceIn);
-            p.fillRect(m_icon.rect(), tintColor);
-            p.end();
-        }
     }
 
     // don't animate initial setting
-    bool animated = (m_animated || m_allowNextAnimation) && !m_oldIcon.isNull() && !m_sizeChanged && !m_blockNextAnimation;
+    bool animated = m_animated && !m_oldIcon.isNull() && !m_sizeChanged && !m_blockNextAnimation && !isSoftwareRendering();
 
     if (animated && m_animation) {
         m_animValue = 0.0;
         m_animation->setStartValue((qreal)0);
         m_animation->setEndValue((qreal)1);
         m_animation->start();
-        m_allowNextAnimation = false;
     } else {
         if (m_animation) {
             m_animation->stop();
@@ -535,11 +480,7 @@ QImage Icon::findIcon(const QSize &size)
             iconSource = QUrl(iconSource).path();
         }
 
-        QIcon icon;
-
-        const QColor tintColor =
-            !m_color.isValid() || m_color == Qt::transparent ? (m_selected ? m_theme->highlightedTextColor() : m_theme->textColor()) : m_color;
-        icon = m_theme->iconFromTheme(iconSource, tintColor);
+        const QIcon icon = loadFromTheme(iconSource);
 
         if (!icon.isNull()) {
             img = iconPixmap(icon);
@@ -554,18 +495,6 @@ QImage Icon::findIcon(const QSize &size)
     return img;
 }
 
-QIcon::Mode Icon::iconMode() const
-{
-    if (!isEnabled()) {
-        return QIcon::Disabled;
-    } else if (m_selected) {
-        return QIcon::Selected;
-    } else if (m_active) {
-        return QIcon::Active;
-    }
-    return QIcon::Normal;
-}
-
 QString Icon::fallback() const
 {
     return m_fallback;
@@ -631,7 +560,33 @@ QSize Icon::iconSizeHint() const
 QImage Icon::iconPixmap(const QIcon &icon) const
 {
     const QSize actualSize = icon.actualSize(iconSizeHint());
-    return icon.pixmap(actualSize, m_devicePixelRatio, iconMode(), QIcon::On).toImage();
+    QIcon sourceIcon = icon;
+
+    // if we have a non-default theme we need to load the icon with
+    // the right colors
+    const QQmlEngine *engine = qmlEngine(this);
+    if (engine && !engine->property("_kirigamiTheme").toString().isEmpty()) {
+        const QString iconName = icon.name();
+        if (!iconName.isEmpty() && QIcon::hasThemeIcon(iconName)) {
+            sourceIcon = loadFromTheme(iconName);
+        }
+    }
+
+    auto mode = QIcon::Mode::Normal;
+    if (isSoftwareRendering()) {
+        if (!isEnabled()) {
+            mode = QIcon::Mode::Disabled;
+        } else if (m_active) {
+            mode = QIcon::Mode::Active;
+        }
+    }
+    return sourceIcon.pixmap(actualSize, m_devicePixelRatio, mode, QIcon::On).toImage();
+}
+
+QIcon Icon::loadFromTheme(const QString &iconName) const
+{
+    const QColor tintColor = !m_color.isValid() || m_color == Qt::transparent ? (m_selected ? m_theme->highlightedTextColor() : m_theme->textColor()) : m_color;
+    return m_theme->iconFromTheme(iconName, tintColor);
 }
 
 void Icon::updatePaintedGeometry()
@@ -754,4 +709,29 @@ void Icon::windowVisibleChanged(bool visible)
     }
 }
 
+QRectF Icon::calculateNodeRect()
+{
+    const QSizeF iconPixSize(m_icon.width() / m_devicePixelRatio, m_icon.height() / m_devicePixelRatio);
+    const QSizeF itemPixSize = QSizeF((size() * m_devicePixelRatio).toSize()) / m_devicePixelRatio;
+    QRectF nodeRect(QPoint(0, 0), itemPixSize);
+
+    if (itemPixSize.width() != 0 && itemPixSize.height() != 0) {
+        if (iconPixSize != itemPixSize) {
+            // At this point, the image will already be scaled, but we need to output it in
+            // the correct aspect ratio, painted centered in the viewport. So:
+            QRectF destination(QPointF(0, 0), QSizeF(m_icon.size()).scaled(m_paintedSize, Qt::KeepAspectRatio));
+            destination.moveCenter(nodeRect.center());
+            destination.moveTopLeft(QPointF(destination.topLeft().toPoint() * m_devicePixelRatio) / m_devicePixelRatio);
+            nodeRect = destination;
+        }
+    }
+
+    return nodeRect;
+}
+
+bool Icon::isSoftwareRendering() const
+{
+    return window() && window()->rendererInterface()->graphicsApi() == QSGRendererInterface::Software;
+}
+
 #include "moc_icon.cpp"
